"""Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""


from collections import OrderedDict
import logging
import sys


class RecordStorage(OrderedDict):
    """Storage of already parsed records are extracted from a trace."""

    def __init__(self, window_size):
        """Default constructor."""
        OrderedDict.__init__(self)

        self.window_size = window_size
        # Mapping from record type to the list of record hashes.
        self.type_record_map = {}

    def get_duration(self):
        """Returns the total duration of stored records."""
        return self.get_max_time() - self.get_min_time() if self.has_records() else 0

    def match_record(self, record):
        """Checks whether the specified collection of ordered by time records contains the specified one."""
        result = None

        if record.__class__.__name__ not in self.type_record_map.keys():
            return None

        record_type_list = self.type_record_map[record.__class__.__name__]

        if not record_type_list:
            return None

        for item_hash in reversed(record_type_list):

            item = self[item_hash]
            if record.get_time() - item.get_time() + 1 > self.window_size:
                break
            elif item.match(record):
                result = item
                break
        return result

    def add_record(self, record):
        """Adds the specified record to the storage."""

        record_hash = record.hash()
        self[record_hash] = record

        if record.__class__.__name__ not in self.type_record_map.keys():
            self.type_record_map[record.__class__.__name__] = []

        logging.debug('Add the record: %s' % record.to_string())
        self.type_record_map[record.__class__.__name__].append(record_hash)

    def remove_record(self, record):
        """Removes the specified record to the storage."""

        record_hash = record.hash()

        record_class_name = record.__class__.__name__

        if record_class_name not in self.type_record_map.keys():
            logging.error('Can\'t remove record of new class: %s' % record_class_name)
            sys.exit(-1)

        class_records = self.type_record_map[record_class_name]

        if not class_records:
            logging.error('No records of such class have been found: %s' % record_class_name)
            sys.exit(-1)

        if record_hash not in class_records:
            logging.error('Can\'t remove non-existing record: %s' % record)
            sys.exit(-1)

        logging.debug('Remove the record: %s' % record.to_string())
        class_records.remove(record_hash)
        del self[record_hash]

    def clear_records_after(self, time):
        """Clear records are stored in this object that have 'time' value is greater or equal than the specified."""
        logging.debug('Clear records after time: %d' % time)
        for record_hash in reversed(self.keys()):
            record = self[record_hash]
            logging.debug('The storage contains the record: %s' % record.to_string())
            if record.get_time() >= time:
                logging.debug('')
                self.remove_record(record)

    def print_all_records(self):
        """Prints the list of records from this storage."""
        for record in self.values():
            logging.info(record.to_string())

    def has_records(self):
        """Check whether this object contains at least one record."""
        return any(self)

    def get_first_record(self):
        """Returns the minimal time-indexed record that this storage contains for now."""
        if not self.type_record_map:
            return None
        return self[next(iter(self))]

    def get_last_record(self):
        """Returns the maximal time-indexed record that this storage contains for now."""
        if not self.type_record_map:
            return None
        return self[next(reversed(self))]

    def get_min_time(self):
        """Returns the minimum value of 'time' attribute for all the records are stored into this object."""
        return self.get_first_record().get_time()

    def get_max_time(self):
        """Returns the maximum value of 'time' attribute for all the records are stored into this object."""
        return self.get_last_record().get_time()

    def empty(self):
        """Check whether the storage is empty."""
        return len(self.keys()) == 0

    def clear(self):
        """Makes the storage clear."""
        logging.debug('Clear the storage')
        self.type_record_map.clear()
        for key in self.keys():
            del self[key]
